/*
 * Copyright 2007,2008 Red Hat, Inc.  All Rights Reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  Redistributions of source code must retain the above copyright notice, this
 *  list of conditions and the following disclaimer.
 *
 *  Redistributions in binary form must reproduce the above copyright notice,
 *  this list of conditions and the following disclaimer in the documentation
 *  and/or other materials provided with the distribution.
 *
 *  Neither the name of Red Hat, Inc. nor the names of its contributors may be
 *  used to endorse or promote products derived from this software without
 *  specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 * 
 * File-opening wrappers for creating correctly-labeled files.  So far, we can
 * assume that this is Linux-specific, so we make many simplifying assumptions.
 */

#include "../../include/autoconf.h"

#ifdef USE_SELINUX
#define USE_SELINUX_UNWRAPPED

#include <k5-label.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <pthread.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <selinux/selinux.h>

/* #define DEBUG 1 */

static pthread_mutex_t labeled_lock = PTHREAD_MUTEX_INITIALIZER;

static security_context_t
push_fscreatecon(const char *pathname, mode_t mode)
{
	security_context_t previous, next;
	const char *fullpath;

	previous = NULL;
	if (is_selinux_enabled()) {
		if (getfscreatecon(&previous) == 0) {
			char *genpath;
			genpath = NULL;
			if (pathname[0] != '/') {
				char *wd;
				size_t len;
				len = 0;
				wd = getcwd(NULL, len);
				if (wd == NULL) {
					if (previous != NULL) {
						freecon(previous);
					}
					return NULL;
				}
				len = strlen(wd) + 1 + strlen(pathname) + 1;
				genpath = malloc(len);
				if (genpath == NULL) {
					free(wd);
					if (previous != NULL) {
						freecon(previous);
					}
					return NULL;
				}
				sprintf(genpath, "%s/%s", wd, pathname);
				free(wd);
				fullpath = genpath;
			} else {
				fullpath = pathname;
			}
			next = NULL;
#ifdef DEBUG
			if (isatty(fileno(stderr))) {
				fprintf(stderr, "Looking up context for "
					"\"%s\"(%05o).\n", fullpath, mode);
			}
#endif
			if (matchpathcon(fullpath, mode, &next) != 0) {
				free(genpath);
				if (previous != NULL) {
					freecon(previous);
				}
				return NULL;
			}
			free(genpath);
#ifdef DEBUG
			if (isatty(fileno(stderr))) {
				fprintf(stderr, "Setting file creation context "
					"to \"%s\".\n", next);
			}
#endif
			if (setfscreatecon(next) != 0) {
				freecon(next);
				if (previous) {
					freecon(previous);
				}
				return NULL;
			}
#ifdef DEBUG
		} else {
			if (isatty(fileno(stderr))) {
				fprintf(stderr, "Unable to determine "
					"current context.\n");
			}
#endif
		}
	}
	return previous;
}

static void
pop_fscreatecon(security_context_t previous)
{
	if (is_selinux_enabled()) {
#ifdef DEBUG
		if (isatty(fileno(stderr))) {
			if (previous != NULL) {
				fprintf(stderr, "Resetting file creation "
					"context to \"%s\".\n", previous);
			} else {
				fprintf(stderr, "Resetting file creation "
					"context to default.\n");
			}
		}
#endif
		setfscreatecon(previous);
		if (previous != NULL) {
			freecon(previous);
		}
	}
}

void *
krb5int_push_fscreatecon_for(const char *pathname)
{
	struct stat st;
	if (stat(pathname, &st) != 0) {
		st.st_mode = S_IRUSR | S_IWUSR;
	}
	return push_fscreatecon(pathname, st.st_mode);
}

void
krb5int_pop_fscreatecon(void *con)
{
	pop_fscreatecon(con);
}

FILE *
krb5int_labeled_fopen(const char *path, const char *mode)
{
	FILE *fp;
	int errno_save;
	security_context_t ctx;

	pthread_mutex_lock(&labeled_lock);
	ctx = push_fscreatecon(path, 0);
	fp = fopen(path, mode);
	errno_save = errno;
	pop_fscreatecon(ctx);
	pthread_mutex_unlock(&labeled_lock);

	errno = errno_save;
	return fp;
}

int
krb5int_labeled_creat(const char *path, mode_t mode)
{
	int fd;
	int errno_save;
	security_context_t ctx;

	pthread_mutex_lock(&labeled_lock);
	ctx = push_fscreatecon(path, 0);
	fd = creat(path, mode);
	errno_save = errno;
	pop_fscreatecon(ctx);
	pthread_mutex_unlock(&labeled_lock);

	errno = errno_save;
	return fd;
}

int
krb5int_labeled_mknod(const char *path, mode_t mode, dev_t dev)
{
	int ret;
	int errno_save;
	security_context_t ctx;

	pthread_mutex_lock(&labeled_lock);
	ctx = push_fscreatecon(path, mode);
	ret = mknod(path, mode, dev);
	errno_save = errno;
	pop_fscreatecon(ctx);
	pthread_mutex_unlock(&labeled_lock);

	errno = errno_save;
	return ret;
}

int
krb5int_labeled_mkdir(const char *path, mode_t mode)
{
	int ret;
	int errno_save;
	security_context_t ctx;

	pthread_mutex_lock(&labeled_lock);
	ctx = push_fscreatecon(path, S_IFDIR);
	ret = mkdir(path, mode);
	errno_save = errno;
	pop_fscreatecon(ctx);
	pthread_mutex_unlock(&labeled_lock);

	errno = errno_save;
	return ret;
}

int
krb5int_labeled_open(const char *path, int flags, ...)
{
	int fd;
	int errno_save;
	security_context_t ctx;
	mode_t mode;
	va_list ap;

	if ((flags & O_CREAT) == 0) {
		return open(path, flags);
	}

	pthread_mutex_lock(&labeled_lock);
	ctx = push_fscreatecon(path, 0);

	va_start(ap, flags);
	mode = va_arg(ap, mode_t);
	fd = open(path, flags, mode);
	va_end(ap);

	errno_save = errno;

	pop_fscreatecon(ctx);
	pthread_mutex_unlock(&labeled_lock);
	return fd;
}

#endif
